# kn_sock/file_transfer.py

import os
import socket
import asyncio
from typing import Optional

CHUNK_SIZE = 4096

# -----------------------------
# 🧱 Common Helpers
# -----------------------------

def _get_filename_from_path(path: str) -> str:
    return os.path.basename(path)

def _send_text(sock: socket.socket, text: str):
    sock.sendall((text + "\n").encode('utf-8'))

def _recv_line(sock: socket.socket) -> str:
    data = b''
    while not data.endswith(b'\n'):
        chunk = sock.recv(1)
        if not chunk:
            break
        data += chunk
    return data.decode('utf-8').strip()

async def _recv_line_async(reader: asyncio.StreamReader) -> str:
    line = await reader.readline()
    return line.decode('utf-8').strip()

# -----------------------------
# 📤 Sync File Sender
# -----------------------------

def send_file(host: str, port: int, filepath: str):
    filename = _get_filename_from_path(filepath)
    filesize = os.path.getsize(filepath)

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.connect((host, port))
        _send_text(sock, filename)
        _send_text(sock, str(filesize))

        print(f"[SYNC] Sending '{filename}' ({filesize} bytes)...")
        with open(filepath, 'rb') as f:
            while (chunk := f.read(CHUNK_SIZE)):
                sock.sendall(chunk)
        print("[SYNC] File sent successfully.")

# -----------------------------
# 📥 Sync File Receiver
# -----------------------------

def start_file_server(port: int, save_dir: str, host: str = '0.0.0.0'):
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.bind((host, port))
    server_socket.listen(5)
    print(f"[SYNC] File server listening on {host}:{port}...")

    while True:
        conn, addr = server_socket.accept()
        print(f"[SYNC] Connection from {addr}")

        with conn:
            filename = _recv_line(conn)
            filesize = int(_recv_line(conn))

            save_path = os.path.join(save_dir, filename)
            print(f"[SYNC] Receiving '{filename}' ({filesize} bytes)...")

            with open(save_path, 'wb') as f:
                remaining = filesize
                while remaining > 0:
                    data = conn.recv(min(CHUNK_SIZE, remaining))
                    if not data:
                        break
                    f.write(data)
                    remaining -= len(data)

            print(f"[SYNC] File saved to {save_path}")

# -----------------------------
# 📤 Async File Sender
# -----------------------------

async def send_file_async(host: str, port: int, filepath: str):
    filename = _get_filename_from_path(filepath)
    filesize = os.path.getsize(filepath)

    reader, writer = await asyncio.open_connection(host, port)
    writer.write((filename + '\n').encode())
    writer.write((str(filesize) + '\n').encode())
    await writer.drain()

    print(f"[ASYNC] Sending '{filename}' ({filesize} bytes)...")
    with open(filepath, 'rb') as f:
        while (chunk := f.read(CHUNK_SIZE)):
            writer.write(chunk)
            await writer.drain()
    writer.close()
    await writer.wait_closed()
    print("[ASYNC] File sent successfully.")

# -----------------------------
# 📥 Async File Receiver
# -----------------------------

async def start_file_server_async(port: int, save_dir: str, host: str = '0.0.0.0'):
    async def handle_client(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        addr = writer.get_extra_info('peername')
        print(f"[ASYNC] Connection from {addr}")

        try:
            filename = await _recv_line_async(reader)
            filesize = int(await _recv_line_async(reader))

            save_path = os.path.join(save_dir, filename)
            print(f"[ASYNC] Receiving '{filename}' ({filesize} bytes)...")

            with open(save_path, 'wb') as f:
                remaining = filesize
                while remaining > 0:
                    data = await reader.read(min(CHUNK_SIZE, remaining))
                    if not data:
                        break
                    f.write(data)
                    remaining -= len(data)

            print(f"[ASYNC] File saved to {save_path}")

        except Exception as e:
            print(f"[ASYNC] Error: {e}")

        writer.close()
        await writer.wait_closed()

    server = await asyncio.start_server(handle_client, host, port)
    print(f"[ASYNC] File server listening on {host}:{port}...")
    async with server:
        await server.serve_forever()



import threading
import time
import socket
import json
import pytest
from kn_sock import send_json

# This is a helper to run a *single-connection* JSON server for testing
def run_single_connection_json_server(port, handler, stop_event):
    """
    Starts a server that accepts only one connection, processes one message,
    then stops (for clean pytest testing).
    """
    srv_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    srv_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    srv_sock.bind(('0.0.0.0', port))
    srv_sock.listen(1)
    srv_sock.settimeout(1)  # 1 second timeout to periodically check stop_event

    try:
        while not stop_event.is_set():
            try:
                client_sock, addr = srv_sock.accept()
            except socket.timeout:
                continue
            with client_sock:
                # Receive data - assuming JSON terminated by newline or fixed protocol
                data = client_sock.recv(4096)
                if not data:
                    continue
                try:
                    decoded = json.loads(data.decode())
                except Exception:
                    # Invalid JSON - ignore or break
                    break
                handler(decoded, addr, client_sock)
                # After handling one message, stop server
                stop_event.set()
    finally:
        srv_sock.close()


@pytest.mark.timeout(5)
def test_handle_json_message_and_server():
    received_data = {}
    stop_event = threading.Event()

    def test_handler(data, addr, client_socket):
        received_data['data'] = data
        received_data['addr'] = addr
        client_socket.sendall(b'{"status": "received"}')

    server_thread = threading.Thread(
        target=run_single_connection_json_server,
        args=(9090, test_handler, stop_event),
        daemon=True
    )
    server_thread.start()

    # Wait a bit for server to start
    time.sleep(0.3)

    test_message = {"message": "Hello, Test!"}
    try:
        response = send_json("localhost", 9090, test_message)
    except Exception as e:
        pytest.fail(f"send_json raised an exception: {e}")

    # Wait for server to process message and stop
    stop_event.wait(timeout=3)
    server_thread.join(timeout=1)

    # Assert server received correct data
    assert 'data' in received_data, "Server handler did not receive any data."
    assert received_data['data'] == test_message, (
        f"Server received incorrect data: expected {test_message}, got {received_data['data']}"
    )
    assert 'addr' in received_data, "Server handler did not capture client address."

    assert isinstance(response, dict), f"Response is not a dict: {response}"
    assert response == {"status": "received"}, (
        f"Unexpected response from server: expected {{'status': 'received'}}, got {response}"
    )

    print("[TEST PASSED] JSON server received and responded correctly.")



import os
import tempfile
import subprocess
import threading
import time
import pytest
import sys
import asyncio
import shutil

from kn_sock import start_ssl_tcp_server, send_ssl_tcp_message, start_async_ssl_tcp_server, send_ssl_tcp_message_async

pytestmark = pytest.mark.skipif(
    not shutil.which('openssl'), reason='openssl not available for test cert generation'
)

def generate_self_signed_cert(tmpdir):
    certfile = os.path.join(tmpdir, 'server.crt')
    keyfile = os.path.join(tmpdir, 'server.key')
    subprocess.check_call([
        'openssl', 'req', '-x509', '-nodes', '-days', '1',
        '-newkey', 'rsa:2048',
        '-keyout', keyfile,
        '-out', certfile,
        '-subj', '/CN=localhost'
    ], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
    return certfile, keyfile

def test_ssl_tcp_sync():
    with tempfile.TemporaryDirectory() as tmpdir:
        certfile, keyfile = generate_self_signed_cert(tmpdir)
        port = 9443
        received = {}
        def handler(data, addr, client_socket):
            received['data'] = data
            client_socket.sendall(b'ACK:' + data)
        server_thread = threading.Thread(
            target=start_ssl_tcp_server,
            args=(port, handler, certfile, keyfile),
            kwargs={'require_client_cert': False},
            daemon=True
        )
        server_thread.start()
        time.sleep(1)
        send_ssl_tcp_message('localhost', port, 'hello ssl', cafile=certfile, verify=False)
        time.sleep(0.5)
        assert received['data'] == b'hello ssl'
        print('[SUCCESS] SSL TCP sync server/client')

@pytest.mark.asyncio
async def test_ssl_tcp_async():
    with tempfile.TemporaryDirectory() as tmpdir:
        certfile, keyfile = generate_self_signed_cert(tmpdir)
        port = 9444
        received = {}
        async def handler(data, addr, writer):
            received['data'] = data
            writer.write(b'ACK:' + data)
            await writer.drain()
        def run_server():
            asyncio.run(start_async_ssl_tcp_server(
                port,
                handler,
                certfile=certfile,
                keyfile=keyfile,
                require_client_cert=False
            ))
        server_thread = threading.Thread(
            target=run_server,
            daemon=True
        )
        server_thread.start()
        await asyncio.sleep(1)
        await send_ssl_tcp_message_async('localhost', port, 'hello ssl', cafile=certfile, verify=False)
        await asyncio.sleep(0.5)
        assert received['data'] == b'hello ssl'
        print('[SUCCESS] SSL TCP async server/client')

def test_tcp_connection_pool_ssl():
    from kn_sock import TCPConnectionPool
    import threading
    import time
    with tempfile.TemporaryDirectory() as tmpdir:
        certfile, keyfile = generate_self_signed_cert(tmpdir)
        port = 9553
        def echo_server():
            def handler(data, addr, client_socket):
                client_socket.sendall(b"ECHO:" + data)
            start_ssl_tcp_server(port, handler, certfile, keyfile, require_client_cert=False)
        server_thread = threading.Thread(target=echo_server, daemon=True)
        server_thread.start()
        time.sleep(1)
        pool = TCPConnectionPool('localhost', port, max_size=2, idle_timeout=5, ssl=True, cafile=certfile, verify=False)
        with pool.connection() as conn:
            conn.sendall(b"ssl pool test")
            data = conn.recv(1024)
            assert data == b"ECHO:ssl pool test"
        pool.closeall()
        print("[SUCCESS] TCPConnectionPool SSL/TLS") 


import os
import socket
import tempfile
from unittest import mock
import pytest

from kn_sock.utils import (
    get_free_port,
    get_local_ip,
    chunked_file_reader,
    recv_all,
    print_progress,
    is_valid_json
)

### get_free_port ###
def test_get_free_port_is_available():
    port = get_free_port()
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("localhost", port))
            print(f"[SUCCESS] get_free_port returned available port: {port}")
    except Exception as e:
        print(f"[FAILED] get_free_port returned unavailable port {port}: {e}")
        raise

### get_local_ip ###
def test_get_local_ip_returns_ip():
    ip = get_local_ip()
    try:
        parts = ip.split(".")
        assert len(parts) == 4
        assert all(0 <= int(part) <= 255 for part in parts)
        print(f"[SUCCESS] get_local_ip returned a valid IP: {ip}")
    except Exception:
        print(f"[FAILED] get_local_ip returned an invalid IP: {ip}")
        raise

### chunked_file_reader ###
def test_chunked_file_reader_reads_in_chunks():
    content = b"abcdefghijklmnopqrstuvwxyz"
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.write(content)
        tmp_path = tmp.name

    try:
        chunks = list(chunked_file_reader(tmp_path, chunk_size=5))
        assert b"".join(chunks) == content
        assert all(len(c) <= 5 for c in chunks)
        print(f"[SUCCESS] chunked_file_reader read all content in chunks.")
    except Exception as e:
        print(f"[FAILED] chunked_file_reader error: {e}")
        raise
    finally:
        os.remove(tmp_path)

### recv_all ###
def test_recv_all_receives_exact_bytes():
    server_sock, client_sock = socket.socketpair()
    test_data = b"hello world"
    client_sock.sendall(test_data)

    try:
        received = recv_all(server_sock, len(test_data))
        assert received == test_data
        print(f"[SUCCESS] recv_all received all data correctly.")
    except Exception as e:
        print(f"[FAILED] recv_all error: {e}")
        raise
    finally:
        server_sock.close()
        client_sock.close()

### print_progress ###
def test_print_progress_displays_correctly(capsys):
    try:
        print_progress(50, 100)
        captured = capsys.readouterr()
        assert "50.00%" in captured.out
        print(f"[SUCCESS] print_progress displayed correctly: {captured.out.strip()}")
    except Exception as e:
        print(f"[FAILED] print_progress error: {e}")
        raise

### is_valid_json ###
def test_is_valid_json_returns_true_for_valid():
    valid = '{"key": "value"}'
    try:
        assert is_valid_json(valid)
        print(f"[SUCCESS] is_valid_json correctly identified valid JSON.")
    except Exception as e:
        print(f"[FAILED] is_valid_json failed on valid JSON: {e}")
        raise

def test_is_valid_json_returns_false_for_invalid():
    invalid = "{invalid json}"
    try:
        assert not is_valid_json(invalid)
        print(f"[SUCCESS] is_valid_json correctly identified invalid JSON.")
    except Exception as e:
        print(f"[FAILED] is_valid_json failed on invalid JSON: {e}")
        raise



import threading
import time
import os
import tempfile
import pytest
from unittest.mock import patch

try:
    import cv2
    import pyaudio
    import numpy as np
    HAS_DEPS = True
except ImportError:
    HAS_DEPS = False

pytestmark = pytest.mark.skipif(not HAS_DEPS, reason='cv2/pyaudio not installed')

from kn_sock.live_stream import LiveStreamServer, LiveStreamClient

def test_live_stream_server_client_init():
    # Create a dummy video file
    with tempfile.NamedTemporaryFile(delete=False, suffix='.mp4') as f:
        f.write(b'\x00' * 1024)  # Not a real video, but enough for instantiation
        video_path = f.name
    try:
        with patch('kn_sock.live_stream.LiveStreamServer._extract_audio', return_value=None):
            server = LiveStreamServer(video_path, video_port=9999, audio_port=10000)
            assert server.video_paths == [video_path]
            client = LiveStreamClient('localhost', video_port=9999, audio_port=10000)
            assert client.host == 'localhost'
            # Just test start/stop do not raise
            server._running.set()
            server.stop()
            client._running.set()
            client.stop()
            print('[SUCCESS] LiveStreamServer/Client instantiation and stop')
    finally:
        os.remove(video_path)

def test_adaptive_bitrate(tmp_path):
    # Use a short dummy wav file and a small video (or mock video)
    video_path = str(tmp_path / "dummy.mp4")
    # Create a dummy video file if not exists (1 black frame)
    import cv2
    import numpy as np
    frame = np.zeros((240, 320, 3), dtype=np.uint8)
    out = cv2.VideoWriter(video_path, cv2.VideoWriter_fourcc(*'mp4v'), 1, (320, 240))
    out.write(frame)
    out.release()
    with patch('kn_sock.live_stream.LiveStreamServer._extract_audio', return_value=None):
        server = LiveStreamServer(video_path, host='127.0.0.1', video_port=9000, audio_port=9001, control_port=9010)
        server_thread = threading.Thread(target=server.start, daemon=True)
        server_thread.start()
        time.sleep(2)
        client = LiveStreamClient('127.0.0.1', 9000, 9001, control_port=9010, video_buffer_ms=100, audio_buffer_ms=100)
        client_thread = threading.Thread(target=client.start, daemon=True)
        client_thread.start()
        time.sleep(5)
        client.stop()
        server.stop()
        # Check that server._client_quality has been updated
        print(f"[TEST] Server client quality settings: {server._client_quality}")
        assert any(q != 80 for q in server._client_quality.values()) or len(server._client_quality) == 0
        print("[SUCCESS] Adaptive bitrate feedback and quality adjustment")

def test_multi_video_selection(tmp_path):
    import cv2
    import numpy as np
    from unittest.mock import patch
    # Create two dummy video files
    video_paths = []
    for i in range(2):
        path = str(tmp_path / f"dummy_{i}.mp4")
        frame = np.full((240, 320, 3), i*100, dtype=np.uint8)  # Different color for each video
        out = cv2.VideoWriter(path, cv2.VideoWriter_fourcc(*'mp4v'), 1, (320, 240))
        out.write(frame)
        out.release()
        video_paths.append(path)
    with patch('kn_sock.live_stream.LiveStreamServer._extract_audio', return_value=None):
        server = LiveStreamServer(video_paths, host='127.0.0.1', video_port=9100, audio_port=9101, control_port=9110)
        server_thread = threading.Thread(target=server.start, daemon=True)
        server_thread.start()
        time.sleep(2)
        # Patch input to select the second video (index 1)
        with patch('builtins.input', return_value='1'):
            client = LiveStreamClient('127.0.0.1', 9100, 9101, control_port=9110, video_buffer_ms=100, audio_buffer_ms=100)
            client_thread = threading.Thread(target=client.start, daemon=True)
            client_thread.start()
            time.sleep(5)
            client.stop()
        server.stop()
        print("[SUCCESS] Multi-video selection test passed.") 


# kn_sock/cli.py

import argparse
import sys
from kn_sock.tcp import send_tcp_message, start_tcp_server, start_ssl_tcp_server, send_ssl_tcp_message
from kn_sock.udp import send_udp_message, start_udp_server
from kn_sock.file_transfer import send_file, start_file_server
from kn_sock.live_stream import start_live_stream, connect_to_live_server
from kn_sock.websocket import start_websocket_server, connect_websocket
from kn_sock.http import http_get, http_post, https_get, https_post, start_http_server
from kn_sock.pubsub import start_pubsub_server, PubSubClient
from kn_sock.rpc import start_rpc_server, RPCClient
import os
import time


def tcp_echo_handler(data, addr, conn):
    print(f"[TCP][SERVER] Received from {addr}: {data}")
    conn.sendall(b"Echo: " + data)


def udp_echo_handler(data, addr, sock):
    print(f"[UDP][SERVER] Received from {addr}: {data.decode()}")
    sock.sendto(b"Echo: " + data, addr)


def run_cli():
    parser = argparse.ArgumentParser(
        description="kn_sock: Simplified socket utilities"
    )

    subparsers = parser.add_subparsers(dest="command", help="Sub-command help")

    # --------------------------
    # send-tcp
    # --------------------------
    tcp_send = subparsers.add_parser("send-tcp", help="Send a message over TCP")
    tcp_send.add_argument("host", type=str, help="Target host")
    tcp_send.add_argument("port", type=int, help="Target port")
    tcp_send.add_argument("message", type=str, help="Message to send")

    # --------------------------
    # send-udp
    # --------------------------
    udp_send = subparsers.add_parser("send-udp", help="Send a message over UDP")
    udp_send.add_argument("host", type=str, help="Target host")
    udp_send.add_argument("port", type=int, help="Target port")
    udp_send.add_argument("message", type=str, help="Message to send")

    # --------------------------
    # send-file
    # --------------------------
    file_send = subparsers.add_parser("send-file", help="Send file over TCP")
    file_send.add_argument("host", type=str, help="Target host")
    file_send.add_argument("port", type=int, help="Target port")
    file_send.add_argument("filepath", type=str, help="Path to file to send")

    # --------------------------
    # run-tcp-server
    # --------------------------
    tcp_server = subparsers.add_parser("run-tcp-server", help="Start a basic TCP echo server")
    tcp_server.add_argument("port", type=int, help="Port to bind server")

    # --------------------------
    # run-udp-server
    # --------------------------
    udp_server = subparsers.add_parser("run-udp-server", help="Start a basic UDP echo server")
    udp_server.add_argument("port", type=int, help="Port to bind server")

    # --------------------------
    # run-file-server
    # --------------------------
    file_server = subparsers.add_parser("run-file-server", help="Start a TCP file receiver")
    file_server.add_argument("port", type=int, help="Port to bind server")
    file_server.add_argument("save_dir", type=str, help="Directory to save received files")

    # --------------------------
    # run-live-server
    # --------------------------
    live_server = subparsers.add_parser("run-live-server", help="Start a live video/audio stream server")
    live_server.add_argument("port", type=int, help="Port for video stream (audio will use port+1 by default)")
    live_server.add_argument("video_paths", type=str, nargs='+', help="Path(s) to video file(s) to stream (one or more)")
    live_server.add_argument("--host", type=str, default="0.0.0.0", help="Host to bind (default: 0.0.0.0)")
    live_server.add_argument("--audio-port", type=int, default=None, help="Port for audio stream (default: port+1)")

    # --------------------------
    # connect-live-server
    # --------------------------
    live_client = subparsers.add_parser("connect-live-server", help="Connect to a live video/audio stream server")
    live_client.add_argument("ip", type=str, help="Server IP address")
    live_client.add_argument("port", type=int, help="Video port (audio will use port+1 by default)")
    live_client.add_argument("--audio-port", type=int, default=None, help="Audio port (default: port+1)")

    # --------------------------
    # run-ssl-tcp-server
    # --------------------------
    ssl_tcp_server = subparsers.add_parser("run-ssl-tcp-server", help="Start a secure SSL/TLS TCP server")
    ssl_tcp_server.add_argument("port", type=int, help="Port to bind server")
    ssl_tcp_server.add_argument("certfile", type=str, help="Path to server certificate (PEM)")
    ssl_tcp_server.add_argument("keyfile", type=str, help="Path to server private key (PEM)")
    ssl_tcp_server.add_argument("--cafile", type=str, default=None, help="CA cert for client cert verification (optional)")
    ssl_tcp_server.add_argument("--require-client-cert", action="store_true", help="Require client certificate (mutual TLS)")
    ssl_tcp_server.add_argument("--host", type=str, default="0.0.0.0", help="Host to bind (default: 0.0.0.0)")

    # --------------------------
    # send-ssl-tcp
    # --------------------------
    ssl_tcp_client = subparsers.add_parser("send-ssl-tcp", help="Send a message over SSL/TLS TCP")
    ssl_tcp_client.add_argument("host", type=str, help="Target host")
    ssl_tcp_client.add_argument("port", type=int, help="Target port")
    ssl_tcp_client.add_argument("message", type=str, help="Message to send")
    ssl_tcp_client.add_argument("--cafile", type=str, default=None, help="CA cert for server verification (optional)")
    ssl_tcp_client.add_argument("--certfile", type=str, default=None, help="Client certificate (PEM) for mutual TLS (optional)")
    ssl_tcp_client.add_argument("--keyfile", type=str, default=None, help="Client private key (PEM) for mutual TLS (optional)")
    ssl_tcp_client.add_argument("--no-verify", action="store_true", help="Disable server certificate verification")

    # --------------------------
    # WebSocket Server
    # --------------------------
    ws_server = subparsers.add_parser("run-websocket-server", help="Start a WebSocket echo server")
    ws_server.add_argument("port", type=int, help="Port to bind server")
    ws_server.add_argument("--host", type=str, default="127.0.0.1", help="Host to bind (default: 127.0.0.1)")

    # WebSocket Client
    ws_client = subparsers.add_parser("websocket-client", help="Connect to a WebSocket server and send a message")
    ws_client.add_argument("host", type=str, help="Server host")
    ws_client.add_argument("port", type=int, help="Server port")
    ws_client.add_argument("message", type=str, help="Message to send")

    # HTTP/HTTPS Client
    http_get_cmd = subparsers.add_parser("http-get", help="HTTP GET request")
    http_get_cmd.add_argument("host", type=str)
    http_get_cmd.add_argument("port", type=int)
    http_get_cmd.add_argument("path", type=str, default="/", nargs="?")
    https_get_cmd = subparsers.add_parser("https-get", help="HTTPS GET request")
    https_get_cmd.add_argument("host", type=str)
    https_get_cmd.add_argument("port", type=int)
    https_get_cmd.add_argument("path", type=str, default="/", nargs="?")
    http_post_cmd = subparsers.add_parser("http-post", help="HTTP POST request")
    http_post_cmd.add_argument("host", type=str)
    http_post_cmd.add_argument("port", type=int)
    http_post_cmd.add_argument("path", type=str, default="/", nargs="?")
    http_post_cmd.add_argument("data", type=str, default="")
    https_post_cmd = subparsers.add_parser("https-post", help="HTTPS POST request")
    https_post_cmd.add_argument("host", type=str)
    https_post_cmd.add_argument("port", type=int)
    https_post_cmd.add_argument("path", type=str, default="/", nargs="?")
    https_post_cmd.add_argument("data", type=str, default="")

    # HTTP Server
    http_server = subparsers.add_parser("run-http-server", help="Start a minimal HTTP server (static + routes)")
    http_server.add_argument("port", type=int)
    http_server.add_argument("--host", type=str, default="127.0.0.1")
    http_server.add_argument("--static-dir", type=str, default=None)

    # PubSub Server
    pubsub_server = subparsers.add_parser("run-pubsub-server", help="Start a pub/sub server")
    pubsub_server.add_argument("port", type=int)
    pubsub_server.add_argument("--host", type=str, default="127.0.0.1")

    # PubSub Client
    pubsub_client = subparsers.add_parser("pubsub-client", help="Connect to pub/sub server, subscribe, publish, and receive")
    pubsub_client.add_argument("host", type=str)
    pubsub_client.add_argument("port", type=int)
    pubsub_client.add_argument("topic", type=str)
    pubsub_client.add_argument("message", type=str, nargs="?", default=None)

    # RPC Server
    rpc_server = subparsers.add_parser("run-rpc-server", help="Start an RPC server with add/echo")
    rpc_server.add_argument("port", type=int)
    rpc_server.add_argument("--host", type=str, default="127.0.0.1")

    # RPC Client
    rpc_client = subparsers.add_parser("rpc-client", help="Connect to RPC server and call a function")
    rpc_client.add_argument("host", type=str)
    rpc_client.add_argument("port", type=int)
    rpc_client.add_argument("function", type=str)
    rpc_client.add_argument("args", nargs="*", help="Arguments for the function")

    # --------------------------
    # Parse args and run
    # --------------------------
    args = parser.parse_args()

    if args.command == "send-tcp":
        send_tcp_message(args.host, args.port, args.message)

    elif args.command == "send-udp":
        send_udp_message(args.host, args.port, args.message)

    elif args.command == "send-file":
        send_file(args.host, args.port, args.filepath)

    elif args.command == "run-tcp-server":
        start_tcp_server(args.port, tcp_echo_handler)

    elif args.command == "run-udp-server":
        start_udp_server(args.port, udp_echo_handler)

    elif args.command == "run-file-server":
        start_file_server(args.port, args.save_dir)

    elif args.command == "run-live-server":
        start_live_stream(args.port, args.video_paths, host=args.host, audio_port=args.audio_port)

    elif args.command == "connect-live-server":
        connect_to_live_server(args.ip, args.port, audio_port=args.audio_port)

    elif args.command == "run-ssl-tcp-server":
        start_ssl_tcp_server(
            args.port,
            tcp_echo_handler,
            certfile=args.certfile,
            keyfile=args.keyfile,
            cafile=args.cafile,
            require_client_cert=args.require_client_cert,
            host=args.host
        )

    elif args.command == "send-ssl-tcp":
        send_ssl_tcp_message(
            args.host,
            args.port,
            args.message,
            cafile=args.cafile,
            certfile=args.certfile,
            keyfile=args.keyfile,
            verify=not args.no_verify
        )

    elif args.command == "run-websocket-server":
        def echo_handler(ws):
            print(f"[WebSocket][SERVER] Client connected: {ws.addr}")
            try:
                while ws.open:
                    msg = ws.recv()
                    if not msg:
                        break
                    print(f"[WebSocket][SERVER] Received: {msg}")
                    ws.send(f"Echo: {msg}")
            finally:
                ws.close()
                print(f"[WebSocket][SERVER] Client disconnected: {ws.addr}")
        import threading
        shutdown_event = threading.Event()
        server_thread = threading.Thread(
            target=start_websocket_server,
            args=(args.host, args.port, echo_handler),
            kwargs={"shutdown_event": shutdown_event},
            daemon=True
        )
        server_thread.start()
        print("[WebSocket][SERVER] Running. Press Ctrl+C to stop.")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            print("[WebSocket][SERVER] Shutting down...")
            shutdown_event.set()
            server_thread.join()

    elif args.command == "websocket-client":
        ws = connect_websocket(args.host, args.port)
        print("[WebSocket][CLIENT] Connected to server.")
        ws.send(args.message)
        print(f"[WebSocket][CLIENT] Sent: {args.message}")
        reply = ws.recv()
        print(f"[WebSocket][CLIENT] Received: {reply}")
        ws.close()
        print("[WebSocket][CLIENT] Connection closed.")

    elif args.command == "http-get":
        body = http_get(args.host, args.port, args.path)
        print(body)
    elif args.command == "https-get":
        body = https_get(args.host, args.port, args.path)
        print(body)
    elif args.command == "http-post":
        body = http_post(args.host, args.port, args.path, data=args.data)
        print(body)
    elif args.command == "https-post":
        body = https_post(args.host, args.port, args.path, data=args.data)
        print(body)

    elif args.command == "run-http-server":
        def hello_route(request, client_sock):
            client_sock.sendall(b"HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\nContent-Length: 5\r\n\r\nHello")
        def echo_post(request, client_sock):
            body = request['raw'].split(b'\r\n\r\n', 1)[-1]
            resp = b"HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\nContent-Length: " + str(len(body)).encode() + b"\r\n\r\n" + body
            client_sock.sendall(resp)
        routes = {
            ("GET", "/hello"): hello_route,
            ("POST", "/echo"): echo_post,
        }
        import threading
        shutdown_event = threading.Event()
        server_thread = threading.Thread(
            target=start_http_server,
            args=(args.host, args.port),
            kwargs={"static_dir": args.static_dir, "routes": routes, "shutdown_event": shutdown_event},
            daemon=True
        )
        server_thread.start()
        print("[HTTP][SERVER] Running. Press Ctrl+C to stop.")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            print("[HTTP][SERVER] Shutting down...")
            shutdown_event.set()
            server_thread.join()

    elif args.command == "run-pubsub-server":
        import threading
        shutdown_event = threading.Event()
        server_thread = threading.Thread(
            target=start_pubsub_server,
            args=(args.port,),
            kwargs={"host": args.host, "shutdown_event": shutdown_event},
            daemon=True
        )
        server_thread.start()
        print("[PubSub][SERVER] Running. Press Ctrl+C to stop.")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            print("[PubSub][SERVER] Shutting down...")
            shutdown_event.set()
            server_thread.join()

    elif args.command == "pubsub-client":
        client = PubSubClient(args.host, args.port)
        client.subscribe(args.topic)
        print(f"[PubSub][CLIENT] Subscribed to '{args.topic}'")
        if args.message:
            client.publish(args.topic, args.message)
            print(f"[PubSub][CLIENT] Published to '{args.topic}': {args.message}")
        print("[PubSub][CLIENT] Waiting for messages (Ctrl+C to exit)...")
        try:
            while True:
                msg = client.recv(timeout=5)
                if msg:
                    print(f"[PubSub][CLIENT] Received: {msg}")
        except KeyboardInterrupt:
            print("[PubSub][CLIENT] Closing connection.")
            client.close()

    elif args.command == "run-rpc-server":
        def add(a, b):
            return a + b
        def echo(msg):
            return msg
        funcs = {"add": add, "echo": echo}
        import threading
        shutdown_event = threading.Event()
        server_thread = threading.Thread(
            target=start_rpc_server,
            args=(args.port, funcs),
            kwargs={"host": args.host, "shutdown_event": shutdown_event},
            daemon=True
        )
        server_thread.start()
        print("[RPC][SERVER] Running. Press Ctrl+C to stop.")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            print("[RPC][SERVER] Shutting down...")
            shutdown_event.set()
            server_thread.join()

    elif args.command == "rpc-client":
        client = RPCClient(args.host, args.port)
        try:
            # Try to parse args as ints or floats if possible
            parsed_args = []
            for a in args.args:
                try:
                    parsed_args.append(int(a))
                except ValueError:
                    try:
                        parsed_args.append(float(a))
                    except ValueError:
                        parsed_args.append(a)
            result = client.call(args.function, *parsed_args)
            print(f"[RPC][CLIENT] {args.function}{tuple(parsed_args)} = {result}")
        except Exception as e:
            print(f"[RPC][CLIENT] Error: {e}")
        finally:
            client.close()

    else:
        parser.print_help()
        sys.exit(1)



import pytest
import time
from kn_sock.decorators import log_exceptions, retry, measure_time, ensure_json_input
from kn_sock.errors import InvalidJSONError

def test_log_exceptions_raises():
    @log_exceptions(raise_error=True)
    def fail():
        raise ValueError('fail!')
    with pytest.raises(ValueError):
        fail()

def test_log_exceptions_no_raise(capsys):
    @log_exceptions(raise_error=False)
    def fail():
        raise ValueError('fail!')
    fail()
    out = capsys.readouterr().out
    assert '[ERROR]' in out

def test_retry_retries(monkeypatch):
    calls = {'n': 0}
    @retry(retries=3, delay=0.01, exceptions=(ValueError,))
    def sometimes_fails():
        calls['n'] += 1
        if calls['n'] < 3:
            raise ValueError('fail')
        return 'ok'
    assert sometimes_fails() == 'ok'
    assert calls['n'] == 3

def test_measure_time(capsys):
    @measure_time
    def slow():
        time.sleep(0.01)
        return 42
    result = slow()
    out = capsys.readouterr().out
    assert '[TIMER]' in out
    assert result == 42

def test_ensure_json_input_accepts_dict():
    @ensure_json_input
    def handler(data):
        return data
    assert handler({'a': 1}) == {'a': 1}

def test_ensure_json_input_accepts_json_str():
    @ensure_json_input
    def handler(data):
        return data
    assert handler('{"a": 1}') == {'a': 1}

def test_ensure_json_input_rejects_invalid():
    @ensure_json_input
    def handler(data):
        return data
    with pytest.raises(InvalidJSONError):
        handler('not json')
    with pytest.raises(InvalidJSONError):
        handler(123) 


import socket
import threading
import time
import os
import queue
import struct
import json
from kn_sock.json_socket import send_json_response, _recv_line

# Dependency checks
try:
    import cv2
    import numpy as np
    import pyaudio
    import wave
    import subprocess
except ImportError as e:
    missing = str(e).split("No module named ")[-1].replace("'", "")
    raise ImportError(f"[kn_sock.live_stream] Missing required package: {missing}. Please install it to use live streaming.")

AUDIO_MAGIC = b'AUD0'

class LiveStreamServer:
    """
    A server that streams video and audio from a file to multiple clients.
    Automatically extracts audio from the video file using FFmpeg.
    """
    def __init__(self, video_paths, host='0.0.0.0', video_port=8000, audio_port=8001, control_port=None):
        if isinstance(video_paths, str):
            video_paths = [video_paths]
        self.video_paths = video_paths
        self.audio_path = "temp_audio.wav"
        self.host = host
        self.video_port = video_port
        self.audio_port = audio_port
        self.control_port = control_port or (video_port + 10)
        self.clients = []
        self.video_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.audio_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.control_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._running = threading.Event()
        self._extract_audio(self.video_paths[0])
        self._client_quality = {}  # addr -> jpeg quality

    def _extract_audio(self, video_path):
        print("[*] Extracting audio from video file...")
        command = [
            'ffmpeg',
            '-i', video_path,
            '-y',
            '-f', 'wav',
            '-ac', '2',
            '-ar', '44100',
            '-vn',
            self.audio_path
        ]
        try:
            subprocess.run(command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
            print(f"[*] Audio extracted successfully to {self.audio_path}")
        except FileNotFoundError:
            print("[!] ERROR: ffmpeg command not found. Please install FFmpeg and ensure it's in your system's PATH.")
            raise
        except subprocess.CalledProcessError as e:
            print(f"[!] ERROR: ffmpeg failed to extract audio. Error:\n{e.stderr.decode()}")
            raise

    def start(self):
        self._running.set()
        self.video_socket.bind((self.host, self.video_port))
        self.video_socket.listen(5)
        print(f"[*] Video server listening on {self.host}:{self.video_port}")
        self.audio_socket.bind((self.host, self.audio_port))
        self.audio_socket.listen(5)
        print(f"[*] Audio server listening on {self.host}:{self.audio_port}")
        self.control_socket.bind((self.host, self.control_port))
        self.control_socket.listen(5)
        print(f"[*] Control server listening on {self.host}:{self.control_port}")
        threading.Thread(target=self._accept_clients, args=(self.video_socket, "video"), daemon=True).start()
        threading.Thread(target=self._accept_clients, args=(self.audio_socket, "audio"), daemon=True).start()
        threading.Thread(target=self._accept_control_clients, daemon=True).start()

    def stop(self):
        self._running.clear()
        for client_socket in self.clients:
            try:
                client_socket.close()
            except socket.error:
                pass
        self.video_socket.close()
        self.audio_socket.close()
        self.control_socket.close()
        try:
            if os.path.exists(self.audio_path):
                os.remove(self.audio_path)
                print(f"[*] Deleted temporary audio file: {self.audio_path}")
        except OSError as e:
            print(f"[!] Error removing temporary audio file: {e}")
        print("[*] Server stopped.")

    def _accept_clients(self, server_socket, stream_type):
        while self._running.is_set():
            try:
                client_socket, addr = server_socket.accept()
                print(f"[*] Accepted {stream_type} connection from {addr[0]}:{addr[1]}")
                self.clients.append(client_socket)
                handler_thread = threading.Thread(target=self._handle_client, args=(client_socket, stream_type), daemon=True)
                handler_thread.start()
            except socket.error:
                break

    def _accept_control_clients(self):
        while self._running.is_set():
            try:
                client_sock, addr = self.control_socket.accept()
                print(f"[*] Accepted control connection from {addr[0]}:{addr[1]}")
                threading.Thread(target=self._handle_control_client, args=(client_sock, addr), daemon=True).start()
            except socket.error:
                break

    def _handle_control_client(self, client_sock, addr):
        # Default quality
        self._client_quality[addr] = 80
        try:
            while self._running.is_set():
                data = b''
                while not data.endswith(b'\n'):
                    chunk = client_sock.recv(1024)
                    if not chunk:
                        break
                    data += chunk
                if not data:
                    break
                try:
                    feedback = json.loads(data.decode())
                    buf_level = feedback.get("buffer_level", 0.2)
                    # Simple logic: if buffer low, reduce quality; if high, increase
                    q = self._client_quality[addr]
                    if buf_level < 0.1 and q > 40:
                        q -= 10
                    elif buf_level > 0.3 and q < 90:
                        q += 10
                    self._client_quality[addr] = max(40, min(90, q))
                except Exception:
                    continue
        finally:
            del self._client_quality[addr]
            client_sock.close()

    def _handle_client(self, client_socket, stream_type):
        try:
            if stream_type == "video":
                # Multi-video handshake
                video_names = [os.path.basename(p) for p in self.video_paths]
                send_json_response(client_socket, {"videos": video_names})
                sel_bytes = _recv_line(client_socket)
                if not sel_bytes:
                    print("[VIDEO] Client disconnected before selection.")
                    return
                try:
                    sel = int(json.loads(sel_bytes.decode().strip()).get("index", 0))
                except Exception:
                    sel = 0
                sel = max(0, min(sel, len(self.video_paths)-1))
                selected_video = self.video_paths[sel]
                print(f"[VIDEO] Client selected video: {selected_video}")
                self._extract_audio(selected_video)
                self._stream_video(client_socket, selected_video)
            elif stream_type == "audio":
                self._stream_audio(client_socket)
        except (ConnectionResetError, BrokenPipeError, ConnectionAbortedError):
            print(f"[*] Client disconnected from {stream_type} stream.")
        finally:
            if client_socket in self.clients:
                self.clients.remove(client_socket)
            client_socket.close()

    def _stream_video(self, client_socket, video_path):
        cap = cv2.VideoCapture(video_path)
        if not cap.isOpened():
            print(f"[!] Could not open video file: {video_path}. Try converting it to mp4 (H.264) format for best compatibility.")
            return
        addr = client_socket.getpeername()
        while self._running.is_set():
            ret, frame = cap.read()
            if not ret:
                cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
                continue
            quality = self._client_quality.get(addr, 80)
            _, buffer = cv2.imencode('.jpg', frame, [cv2.IMWRITE_JPEG_QUALITY, quality])
            data = np.array(buffer).tobytes()
            timestamp = time.time()
            try:
                client_socket.sendall(struct.pack('!dI', timestamp, len(data)) + data)
            except (socket.error, BrokenPipeError):
                break
            time.sleep(1/30)
        cap.release()

    def _stream_audio(self, client_socket):
        try:
            with wave.open(self.audio_path, 'rb') as wf:
                chunk_size = 1024
                while self._running.is_set():
                    data = wf.readframes(chunk_size)
                    if not data:
                        wf.rewind()
                        data = wf.readframes(chunk_size)
                    timestamp = time.time()
                    try:
                        # [4 bytes magic][8 bytes timestamp][4 bytes chunk_len][chunk]
                        header = AUDIO_MAGIC + struct.pack('!dI', timestamp, len(data))
                        client_socket.sendall(header + data)
                    except (socket.error, BrokenPipeError):
                        break
        except FileNotFoundError:
            return

def start_live_stream(port, video_paths, host='0.0.0.0', audio_port=None):
    """
    Starts a live stream server for the given video file(s).
    Args:
        port (int): Port for video stream.
        video_paths (list[str]): Path(s) to video file(s).
        host (str): Host to bind (default 0.0.0.0).
        audio_port (int): Port for audio stream (default: port+1).
    """
    if audio_port is None:
        audio_port = port + 1
    server = LiveStreamServer(video_paths, host, port, audio_port)
    try:
        server.start()
        print("[kn_sock] Live stream server started. Press Ctrl+C to stop.")
        while True:
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        print("\n[kn_sock] Stopping live stream server...")
    finally:
        server.stop()

class JitterBuffer:
    """
    Thread-safe buffer to smooth out irregular frame/chunk arrival times.
    Use for video/audio streaming to reduce jitter/stutter.
    """
    def __init__(self, max_delay=0.2, target_interval=1/30):
        self.buffer = queue.Queue()
        self.max_delay = max_delay  # seconds of buffer (e.g., 0.2s)
        self.target_interval = target_interval  # e.g., 1/30 for 30fps
        self._running = threading.Event()

    def start(self, process_func):
        self._running.set()
        threading.Thread(target=self._playback_loop, args=(process_func,), daemon=True).start()

    def stop(self):
        self._running.clear()
        with self.buffer.mutex:
            self.buffer.queue.clear()

    def put(self, item, timestamp=None):
        if timestamp is None:
            timestamp = time.time()
        self.buffer.put((timestamp, item))

    def _playback_loop(self, process_func):
        # Wait until buffer is filled to max_delay
        base_ts = None
        debug_count = 0
        while self._running.is_set() and self.buffer.qsize() * self.target_interval < self.max_delay:
            time.sleep(self.target_interval / 2)
        while self._running.is_set():
            try:
                ts, item = self.buffer.get(timeout=self.target_interval)
                if base_ts is None:
                    base_ts = ts
                    start_play = time.time()
                # Calculate when to play this item
                now = time.time()
                play_at = start_play + (ts - base_ts)
                delay = play_at - now
                # Sanity check: skip if delay is absurd
                if delay > 2 or delay < -2:
                    print(f"[JitterBuffer] Skipping frame/chunk due to unreasonable delay: {delay:.3f}s (ts={ts}, base_ts={base_ts})")
                    continue
                if delay > 0:
                    time.sleep(delay)
                process_func(item)
            except queue.Empty:
                continue

class LiveStreamClient:
    """
    A client that receives and plays back video and audio streams from a server.
    Now uses a JitterBuffer for smooth playback.
    """
    def __init__(self, host='127.0.0.1', video_port=8000, audio_port=8001, video_buffer_ms=200, audio_buffer_ms=200, video_fps=30, control_port=None):
        self.host = host
        self.video_port = video_port
        self.audio_port = audio_port
        self.control_port = control_port or (video_port + 10)
        self.video_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.audio_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.control_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._running = threading.Event()
        self.p = pyaudio.PyAudio()
        self.stream = self.p.open(format=pyaudio.paInt16,
                                  channels=2,
                                  rate=44100,
                                  output=True,
                                  frames_per_buffer=1024)
        self.video_jitter = JitterBuffer(max_delay=video_buffer_ms/1000, target_interval=1/video_fps)
        self.audio_jitter = JitterBuffer(max_delay=audio_buffer_ms/1000, target_interval=1024/44100)

    def start(self):
        self._running.set()
        try:
            self.video_socket.connect((self.host, self.video_port))
            print("[*] Connected to video stream.")
            # Multi-video handshake
            video_list_bytes = _recv_line(self.video_socket)
            video_list = json.loads(video_list_bytes.decode().strip())
            videos = video_list.get("videos", [])
            if len(videos) > 1:
                print("Available videos:")
                for i, v in enumerate(videos):
                    print(f"  {i}: {v}")
                while True:
                    try:
                        sel = int(input(f"Select video [0-{len(videos)-1}]: "))
                        if 0 <= sel < len(videos):
                            break
                    except Exception:
                        pass
                sel_json = json.dumps({"index": sel}) + '\n'
                self.video_socket.sendall(sel_json.encode('utf-8'))
            else:
                self.video_socket.sendall(json.dumps({"index": 0}).encode('utf-8') + b'\n')
            self.audio_socket.connect((self.host, self.audio_port))
            print("[*] Connected to audio stream.")
            self.control_socket.connect((self.host, self.control_port))
            print("[*] Connected to control channel.")
        except ConnectionRefusedError:
            print("[!] Connection refused. Make sure the server is running.")
            return
        self.video_jitter.start(self._play_video_frame)
        self.audio_jitter.start(self._play_audio_chunk)
        threading.Thread(target=self._send_feedback_loop, daemon=True).start()
        video_thread = threading.Thread(target=self._receive_video)
        audio_thread = threading.Thread(target=self._receive_audio)
        video_thread.start()
        audio_thread.start()

    def stop(self):
        self._running.clear()
        try:
            self.video_socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        try:
            self.audio_socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        try:
            self.control_socket.shutdown(socket.SHUT_RDWR)
        except Exception:
            pass
        self.video_socket.close()
        self.audio_socket.close()
        self.control_socket.close()
        self.stream.stop_stream()
        self.stream.close()
        self.p.terminate()
        self.video_jitter.stop()
        self.audio_jitter.stop()
        try:
            import cv2
            cv2.destroyAllWindows()
        except Exception:
            pass
        print("[*] Client stopped.")

    def _play_video_frame(self, frame):
        import cv2
        cv2.imshow('Live Stream', frame)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            self.stop()

    def _play_audio_chunk(self, chunk):
        self.stream.write(chunk)

    def _send_feedback_loop(self):
        while self._running.is_set():
            # Feedback: buffer fill level (in seconds)
            buf_level = self.video_jitter.buffer.qsize() * self.video_jitter.target_interval
            msg = json.dumps({"buffer_level": buf_level}) + '\n'
            try:
                self.control_socket.sendall(msg.encode())
            except Exception:
                break
            time.sleep(1)

    def _receive_video(self):
        import cv2
        data = b''
        payload_size = 12  # 8 bytes timestamp + 4 bytes length
        while self._running.is_set():
            try:
                while len(data) < payload_size:
                    packet = self.video_socket.recv(4 * 1024)
                    if not packet:
                        break
                    data += packet
                if not data:
                    break
                packed = data[:payload_size]
                data = data[payload_size:]
                timestamp, msg_size = struct.unpack('!dI', packed)
                while len(data) < msg_size:
                    data += self.video_socket.recv(4 * 1024)
                frame_data = data[:msg_size]
                data = data[msg_size:]
                frame = cv2.imdecode(np.frombuffer(frame_data, np.uint8), cv2.IMREAD_COLOR)
                self.video_jitter.put(frame, timestamp=timestamp)
            except (ConnectionResetError, BrokenPipeError):
                print("[!] Lost connection to video stream.")
                break
        self.stop()

    def _receive_audio(self):
        chunk_size = 1024
        ts_size = 8
        magic_size = 4
        len_size = 4
        data = b''
        debug_count = 0
        while self._running.is_set():
            try:
                # Find magic number
                while len(data) < magic_size:
                    packet = self.audio_socket.recv(magic_size - len(data))
                    if not packet:
                        print("[AUDIO] Socket closed while reading magic.")
                        return
                    data += packet
                if data[:magic_size] != AUDIO_MAGIC:
                    # Shift by one and try again
                    data = data[1:]
                    continue
                data = data[magic_size:]
                # Read timestamp and chunk_len
                while len(data) < ts_size + len_size:
                    packet = self.audio_socket.recv(ts_size + len_size - len(data))
                    if not packet:
                        print("[AUDIO] Socket closed while reading header.")
                        return
                    data += packet
                # timestamp = struct.unpack('!d', data[:ts_size])[0]  # Ignore timestamp for audio
                chunk_len = struct.unpack('!I', data[ts_size:ts_size+len_size])[0]
                data = data[ts_size+len_size:]
                # Read chunk
                while len(data) < chunk_len:
                    packet = self.audio_socket.recv(chunk_len - len(data))
                    if not packet:
                        print("[AUDIO] Socket closed while reading chunk.")
                        return
                    data += packet
                chunk = data[:chunk_len]
                data = data[chunk_len:]
                if debug_count < 5:
                    print(f"[AUDIO][DEBUG] chunk_size={len(chunk)}")
                    debug_count += 1
                self.stream.write(chunk)
            except (ConnectionResetError, BrokenPipeError):
                print("[!] Lost connection to audio stream.")
                break
            except Exception as e:
                print(f"[AUDIO][ERROR] {e}")
                break

def connect_to_live_server(ip, port, audio_port=None):
    """
    Connects to a live stream server and plays the video/audio.
    Args:
        ip (str): Server IP address.
        port (int): Video port.
        audio_port (int): Audio port (default: port+1).
    """
    if audio_port is None:
        audio_port = port + 1
    client = LiveStreamClient(ip, port, audio_port)
    try:
        client.start()
        print("[kn_sock] Connected to live stream. Press 'q' in the video window or Ctrl+C to stop.")
        while client._running.is_set():
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        print("\n[kn_sock] Stopping live stream client...")
    finally:
        client.stop() 


import pytest

from kn_sock.errors import (
    EasySocketError,
    ConnectionTimeoutError,
    PortInUseError,
    InvalidJSONError,
    UnsupportedProtocolError,
    FileTransferError,
)

### EasySocketError ###
def test_easy_socket_error():
    try:
        raise EasySocketError("General socket error")
    except EasySocketError as e:
        assert isinstance(e, EasySocketError)
        assert str(e) == "General socket error"
        print("[SUCCESS] EasySocketError raised and caught successfully.")

### ConnectionTimeoutError ###
def test_connection_timeout_error():
    try:
        raise ConnectionTimeoutError("Connection timed out")
    except ConnectionTimeoutError as e:
        assert isinstance(e, ConnectionTimeoutError)
        assert str(e) == "Connection timed out"
        print("[SUCCESS] ConnectionTimeoutError raised and caught successfully.")

### PortInUseError ###
def test_port_in_use_error():
    try:
        raise PortInUseError(5000)
    except PortInUseError as e:
        assert isinstance(e, PortInUseError)
        assert str(e) == "Port 5000 is already in use."
        print("[SUCCESS] PortInUseError raised and caught successfully.")

### UnsupportedProtocolError ###
def test_unsupported_protocol_error():
    try:
        raise UnsupportedProtocolError("XYZ")
    except UnsupportedProtocolError as e:
        assert isinstance(e, UnsupportedProtocolError)
        assert str(e) == "Protocol 'XYZ' is not supported."
        print("[SUCCESS] UnsupportedProtocolError raised and caught successfully.")

### InvalidJSONError ###
def test_invalid_json_error():
    try:
        raise InvalidJSONError("Invalid JSON data")
    except InvalidJSONError as e:
        assert isinstance(e, InvalidJSONError)
        assert str(e) == "Invalid JSON data"
        print("[SUCCESS] InvalidJSONError raised and caught successfully.")


### FileTransferError ###
def test_file_transfer_error():
    try:
        raise FileTransferError("File transfer failed")
    except FileTransferError as e:
        assert isinstance(e, FileTransferError)
        assert str(e) == "File transfer failed"
        print("[SUCCESS] FileTransferError raised and caught successfully.")



import os
import tempfile
import threading
import asyncio
import pytest
from kn_sock.file_transfer import send_file, start_file_server, send_file_async, start_file_server_async
from kn_sock.utils import get_free_port

@pytest.fixture
def temp_text_file():
    with tempfile.NamedTemporaryFile(delete=False, mode='w', suffix='.txt') as f:
        f.write('hello world')
        f.flush()
        yield f.name
    os.remove(f.name)

@pytest.fixture
def temp_dir():
    d = tempfile.mkdtemp()
    yield d
    try:
        os.rmdir(d)
    except Exception:
        pass

def test_sync_file_transfer(temp_text_file, temp_dir):
    port = get_free_port()
    received_path = os.path.join(temp_dir, os.path.basename(temp_text_file))
    def server():
        start_file_server(port, temp_dir)
    server_thread = threading.Thread(target=server, daemon=True)
    server_thread.start()
    import time; time.sleep(0.5)
    send_file('localhost', port, temp_text_file)
    import time; time.sleep(0.5)
    assert os.path.exists(received_path)
    with open(received_path) as f:
        assert f.read() == 'hello world'
    print('[SUCCESS] Sync file transfer')

@pytest.mark.asyncio
async def test_async_file_transfer(temp_text_file, temp_dir):
    port = get_free_port()
    received_path = os.path.join(temp_dir, os.path.basename(temp_text_file))
    async def server():
        await start_file_server_async(port, temp_dir)
    server_task = asyncio.create_task(server())
    await asyncio.sleep(0.5)
    await send_file_async('localhost', port, temp_text_file)
    await asyncio.sleep(0.5)
    assert os.path.exists(received_path)
    with open(received_path) as f:
        assert f.read() == 'hello world'
    server_task.cancel()
    try:
        await server_task
    except asyncio.CancelledError:
        pass
    print('[SUCCESS] Async file transfer') 


# kn_sock/__init__.py

"""
kn_sock
-----------
A simplified socket programming toolkit for Python.

Features:
- TCP/UDP messaging (sync & async)
- JSON socket communication
- File transfer over TCP
- Threaded/multi-client support
- Command-line interface
"""

__version__ = "0.1.0"
__author__ = "Khagendra Neupane"
__license__ = "MIT"

# TCP
from .tcp import (
    send_tcp_message,
    start_tcp_server,
    start_threaded_tcp_server,
    start_async_tcp_server,
    send_tcp_message_async,
    start_ssl_tcp_server,
    send_ssl_tcp_message,
    start_async_ssl_tcp_server,
    send_ssl_tcp_message_async,
    TCPConnectionPool
)

# UDP
from .udp import (
    send_udp_message,
    start_udp_server,
    send_udp_message_async,
    start_udp_server_async,
    send_udp_multicast,
    start_udp_multicast_server
)

# File Transfer
from .file_transfer import (
    send_file,
    start_file_server,
    send_file_async,
    start_file_server_async
)

# JSON Socket
from .json_socket import (
    start_json_server,
    send_json,
    start_json_server_async,
    send_json_async,
    send_json_response,
    send_json_response_async
)

# Utilities & Errors
from . import utils
from . import errors

# Live Stream
from .live_stream import start_live_stream, connect_to_live_server
from .websocket import start_websocket_server, connect_websocket
from .http import http_get, http_post, https_get, https_post, start_http_server
from .pubsub import start_pubsub_server, PubSubClient
from .rpc import start_rpc_server, RPCClient



import socket
import threading
import json
from typing import Callable, Dict, Set, Optional

class PubSubServer:
    def __init__(self):
        self.topics: Dict[str, Set[socket.socket]] = {}
        self.lock = threading.Lock()

    def subscribe(self, topic: str, client_sock: socket.socket):
        with self.lock:
            self.topics.setdefault(topic, set()).add(client_sock)

    def unsubscribe(self, topic: str, client_sock: socket.socket):
        with self.lock:
            if topic in self.topics:
                self.topics[topic].discard(client_sock)
                if not self.topics[topic]:
                    del self.topics[topic]

    def publish(self, topic: str, message: str):
        with self.lock:
            for sock in list(self.topics.get(topic, [])):
                try:
                    data = json.dumps({"topic": topic, "message": message}).encode() + b"\n"
                    sock.sendall(data)
                except Exception:
                    self.topics[topic].discard(sock)

    def remove_client(self, client_sock: socket.socket):
        with self.lock:
            for topic in list(self.topics):
                self.topics[topic].discard(client_sock)
                if not self.topics[topic]:
                    del self.topics[topic]

def start_pubsub_server(
    port: int,
    handler_func: Optional[Callable[[dict, socket.socket, PubSubServer], None]] = None,
    host: str = '0.0.0.0',
    shutdown_event: Optional[threading.Event] = None
):
    """
    Start a TCP pub/sub server. Handles subscribe, unsubscribe, and publish actions.
    Args:
        port (int): Port to bind.
        handler_func (callable, optional): Custom handler for messages (default: built-in).
        host (str): Host to bind (default '0.0.0.0').
        shutdown_event (threading.Event, optional): For graceful shutdown.
    """
    server = PubSubServer()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(5)
    print(f"[PubSub][SERVER] Listening on {host}:{port}")
    def client_thread(client_sock, addr):
        try:
            buffer = b''
            while True:
                chunk = client_sock.recv(4096)
                if not chunk:
                    break
                buffer += chunk
                while b'\n' in buffer:
                    line, buffer = buffer.split(b'\n', 1)
                    try:
                        msg = json.loads(line.decode())
                    except Exception:
                        continue
                    if handler_func:
                        handler_func(msg, client_sock, server)
                    else:
                        action = msg.get('action')
                        topic = msg.get('topic')
                        if action == 'subscribe':
                            server.subscribe(topic, client_sock)
                        elif action == 'unsubscribe':
                            server.unsubscribe(topic, client_sock)
                        elif action == 'publish':
                            server.publish(topic, msg.get('message', ''))
        finally:
            server.remove_client(client_sock)
            client_sock.close()
    try:
        while True:
            if shutdown_event is not None and shutdown_event.is_set():
                print("[PubSub][SERVER] Shutdown event set. Stopping server.")
                break
            sock.settimeout(1.0)
            try:
                client_sock, addr = sock.accept()
            except socket.timeout:
                continue
            threading.Thread(target=client_thread, args=(client_sock, addr), daemon=True).start()
    finally:
        sock.close()
        print("[PubSub][SERVER] Shutdown complete.")

class PubSubClient:
    """
    Simple TCP pub/sub client. Use subscribe, unsubscribe, publish, and recv methods.
    """
    def __init__(self, host: str, port: int):
        self.sock = socket.create_connection((host, port))
        self.lock = threading.Lock()
        self.recv_buffer = b''

    def subscribe(self, topic: str):
        self._send({"action": "subscribe", "topic": topic})

    def unsubscribe(self, topic: str):
        self._send({"action": "unsubscribe", "topic": topic})

    def publish(self, topic: str, message: str):
        self._send({"action": "publish", "topic": topic, "message": message})

    def recv(self, timeout: float = None) -> Optional[dict]:
        self.sock.settimeout(timeout)
        while True:
            if b'\n' in self.recv_buffer:
                line, self.recv_buffer = self.recv_buffer.split(b'\n', 1)
                try:
                    return json.loads(line.decode())
                except Exception:
                    continue
            try:
                chunk = self.sock.recv(4096)
                if not chunk:
                    return None
                self.recv_buffer += chunk
            except socket.timeout:
                return None

    def _send(self, msg: dict):
        with self.lock:
            self.sock.sendall(json.dumps(msg).encode() + b'\n')

    def close(self):
        self.sock.close() 


from setuptools import setup, find_packages

setup(
    name="kn-sock",
    version="0.2.4",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'kn-sock=kn_sock.cli:run_cli',
        ],
    },
    install_requires=[
        "opencv-python",
        "numpy",
        "pyaudio",
        "ffmpeg-python"
    ],
    author="Khagendra Neupane",
    author_email="nkhagendra1@gmail.com",
    description="A simplified socket programming toolkit for Python",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/KhagendraN/kn-sock",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



# kn_sock/errors.py

class EasySocketError(Exception):
    """Base exception for all easy_socket errors."""
    pass

# -----------------------------
# Connection-related errors
# -----------------------------

class ConnectionTimeoutError(EasySocketError):
    """Raised when a connection or read/write operation times out."""
    def __init__(self, message="Connection timed out."):
        super().__init__(message)

class PortInUseError(EasySocketError):
    """Raised when a specified port is already in use."""
    def __init__(self, port, message=None):
        message = message or f"Port {port} is already in use."
        super().__init__(message)

# -----------------------------
# Data & Protocol errors
# -----------------------------

class InvalidJSONError(EasySocketError):
    """Raised when a JSON message cannot be decoded."""
    def __init__(self, message="Received invalid JSON data."):
        super().__init__(message)

class UnsupportedProtocolError(EasySocketError):
    """Raised when a requested protocol is not supported."""
    def __init__(self, protocol, message=None):
        message = message or f"Protocol '{protocol}' is not supported."
        super().__init__(message)

# -----------------------------
# File transfer errors
# -----------------------------

class FileTransferError(EasySocketError):
    """Raised when file transfer fails."""
    def __init__(self, message="File transfer failed."):
        super().__init__(message)



import socket
import threading
import base64
import hashlib
import struct
from typing import Callable

GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

class WebSocketConnection:
    def __init__(self, conn, addr):
        self.conn = conn
        self.addr = addr
        self.open = True

    def send(self, message: str):
        # Send a text frame
        payload = message.encode('utf-8')
        header = b'\x81'  # FIN + text frame
        length = len(payload)
        if length < 126:
            header += struct.pack('B', length)
        elif length < (1 << 16):
            header += struct.pack('!BH', 126, length)
        else:
            header += struct.pack('!BQ', 127, length)
        self.conn.sendall(header + payload)

    def recv(self) -> str:
        # Receive a text frame (no fragmentation, no extensions)
        first2 = self.conn.recv(2)
        if not first2:
            self.open = False
            return ''
        fin_opcode, mask_len = first2
        masked = mask_len & 0x80
        length = mask_len & 0x7F
        if length == 126:
            length = struct.unpack('!H', self.conn.recv(2))[0]
        elif length == 127:
            length = struct.unpack('!Q', self.conn.recv(8))[0]
        if masked:
            mask = self.conn.recv(4)
            data = bytearray(self.conn.recv(length))
            for i in range(length):
                data[i] ^= mask[i % 4]
            return data.decode('utf-8')
        else:
            data = self.conn.recv(length)
            return data.decode('utf-8')

    def close(self):
        # Send close frame
        try:
            self.conn.sendall(b'\x88\x00')
        except Exception:
            pass
        self.conn.close()
        self.open = False


def _handshake(conn):
    # Minimal WebSocket handshake
    request = b''
    while b'\r\n\r\n' not in request:
        chunk = conn.recv(1024)
        if not chunk:
            return False
        request += chunk
    headers = {}
    for line in request.decode().split('\r\n')[1:]:
        if ': ' in line:
            k, v = line.split(': ', 1)
            headers[k.lower()] = v
    key = headers.get('sec-websocket-key')
    if not key:
        return False
    accept = base64.b64encode(hashlib.sha1((key + GUID).encode()).digest()).decode()
    response = (
        'HTTP/1.1 101 Switching Protocols\r\n'
        'Upgrade: websocket\r\n'
        'Connection: Upgrade\r\n'
        f'Sec-WebSocket-Accept: {accept}\r\n\r\n'
    )
    conn.sendall(response.encode())
    return True


def start_websocket_server(host: str, port: int, handler: Callable[[WebSocketConnection], None], shutdown_event=None):
    """
    Start a minimal WebSocket server.
    Args:
        host (str): Host to bind.
        port (int): Port to bind.
        handler (callable): Function called with WebSocketConnection for each client.
        shutdown_event (threading.Event, optional): For graceful shutdown.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(5)
    print(f"[WebSocket][SERVER] Listening on {host}:{port}")
    try:
        while True:
            if shutdown_event is not None and shutdown_event.is_set():
                print("[WebSocket][SERVER] Shutdown event set. Stopping server.")
                break
            sock.settimeout(1.0)
            try:
                conn, addr = sock.accept()
            except socket.timeout:
                continue
            if not _handshake(conn):
                conn.close()
                continue
            ws = WebSocketConnection(conn, addr)
            threading.Thread(target=handler, args=(ws,), daemon=True).start()
    finally:
        sock.close()
        print("[WebSocket][SERVER] Shutdown complete.")


def connect_websocket(host: str, port: int, resource: str = '/', headers: dict = None) -> WebSocketConnection:
    """
    Connect to a WebSocket server and return a WebSocketConnection.
    Args:
        host (str): Server host.
        port (int): Server port.
        resource (str): Resource path (default '/').
        headers (dict): Additional headers.
    Returns:
        WebSocketConnection
    """
    import os
    import random
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    key = base64.b64encode(os.urandom(16)).decode()
    req = (
        f"GET {resource} HTTP/1.1\r\n"
        f"Host: {host}:{port}\r\n"
        "Upgrade: websocket\r\n"
        "Connection: Upgrade\r\n"
        f"Sec-WebSocket-Key: {key}\r\n"
        "Sec-WebSocket-Version: 13\r\n"
    )
    if headers:
        for k, v in headers.items():
            req += f"{k}: {v}\r\n"
    req += "\r\n"
    sock.sendall(req.encode())
    # Read response
    resp = b''
    while b'\r\n\r\n' not in resp:
        chunk = sock.recv(1024)
        if not chunk:
            raise ConnectionError("WebSocket handshake failed")
        resp += chunk
    if b"101" not in resp.split(b"\r\n", 1)[0]:
        raise ConnectionError("WebSocket handshake failed")
    return WebSocketConnection(sock, (host, port)) 


# kn_sock/decorators.py

import time
import json
import functools
import traceback
from typing import Callable, Any, Optional, Type
from kn_sock.errors import InvalidJSONError


# -----------------------------
# 🧾 Log Exceptions
# -----------------------------

def log_exceptions(raise_error: bool = True):
    """
    Logs exceptions and optionally re-raises them.
    """
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                print(f"[ERROR] Exception in '{func.__name__}': {e}")
                traceback.print_exc()
                if raise_error:
                    raise
        return wrapper
    return decorator


# -----------------------------
# 🔁 Retry Decorator
# -----------------------------

def retry(
    retries: int = 3,
    delay: float = 1.0,
    exceptions: tuple = (Exception,)
):
    """
    Retries a function upon failure, with delay.
    """
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(1, retries + 1):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    print(f"[RETRY] {func.__name__}: Attempt {attempt} failed: {e}")
                    if attempt < retries:
                        time.sleep(delay)
                    else:
                        raise
        return wrapper
    return decorator


# -----------------------------
# ⏱️ Measure Execution Time
# -----------------------------

def measure_time(func: Callable):
    """
    Measures and prints execution time of the wrapped function.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        elapsed = time.time() - start
        print(f"[TIMER] {func.__name__} executed in {elapsed:.4f} seconds")
        return result
    return wrapper


# -----------------------------
# ✅ Ensure JSON Input
# -----------------------------

def ensure_json_input(func: Callable):
    """
    Validates that the first argument is a valid JSON object (dict or str).
    Raises InvalidJSONError otherwise.
    """
    @functools.wraps(func)
    def wrapper(data, *args, **kwargs):
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except json.JSONDecodeError:
                raise InvalidJSONError("Handler received invalid JSON string.")
        elif not isinstance(data, dict):
            raise InvalidJSONError("Handler expects JSON object or string.")

        return func(data, *args, **kwargs)
    return wrapper
